// (c) 2002 by Jrn Loviscach

#include <math.h>
#include <assert.h>

#include "bezierpath.h"
#include "adapt.h"

BezierPath::BezierPath()
: bezierVector(0), hasPath(false)
{}

BezierPath::~BezierPath()
{
	if(hasPath)
	{
		for(int j=0; j<bezierVector->size(); ++j)
		{
			delete (*bezierVector)[j];
		}
		DELETE(bezierVector);
	}
}

void BezierPath::Init(const PixelPath& pp)
{
	assert(!hasPath);
	bezierVector = NEW(std::vector<Bezier*>);
	hasPath = true;

	int len = pp.GetLength();
	assert(len > 1);  // need two pixels at least

	double* listX = BNEW(double, len);
	double* listY = BNEW(double, len);
	double* listInvD = BNEW(double, len);

	int i;
	for(i=0; i<len; i++)
	{
		listX[i] = pp.GetX(i);
		listY[i] = pp.GetY(i);
		listInvD[i] = pp.GetInvD(i);
	}

	int j;
	for(j=0; j<20; j++) // first do several rounds of smoothing
	{
		double* nListX = BNEW(double, len);
		double* nListY = BNEW(double, len);
		double* nListInvD = BNEW(double, len);

		nListX[0] = (1.0/3.0)*(2.0*listX[0] + listX[1]);
		nListY[0] = (1.0/3.0)*(2.0*listY[0] + listY[1]);
		nListInvD[0] = (1.0/3.0)*(2.0*listInvD[0] + listInvD[1]);
		nListX[len-1] = (1.0/3.0)*(2.0*listX[len-1] + listX[len-1]);
		nListY[len-1] = (1.0/3.0)*(2.0*listY[len-1] + listY[len-1]);
		nListInvD[len-1] = (1.0/3.0)*(2.0*listInvD[len-1] + listInvD[len-1]);

		for(i=1; i<len-1; i++)
		{
			// Is there a sharp corner at i?
			// calculate the mean direction of the two (or one) vertices before
			// and the two (or one) behind
			double dx1 = listX[i] - listX[(i>1) ? i-2 : i-1];
			double dx2 = listX[(i<len-2) ? i+2 : i+1] - listX[i];
			double dy1 = listY[i] - listY[(i>1) ? i-2 : i-1];
			double dy2 = listY[(i<len-2) ? i+2 : i+1] - listY[i];
			double cosPhi = (dx1*dx2 + dy1*dy2)/(sqrt(dx1*dx1+dy1*dy1)*sqrt(dx2*dx2+dy2*dy2)+1e-50);
			// possible extensions: check for deviation of invD
			if(cosPhi > cos(45.0*(3.14/180.0))) // quite a straight line
			{
				nListX[i] = 0.25*(1.0*listX[i-1] + 1.0*listX[i+1] + 2.0*listX[i]);
				nListY[i] = 0.25*(1.0*listY[i-1] + 1.0*listY[i+1] + 2.0*listY[i]);
				nListInvD[i] = 0.25*(1.0*listInvD[i-1] + 1.0*listInvD[i+1] + 2.0*listInvD[i]);
			}
			else
			{
				nListX[i] = listX[i];
				nListY[i] = listY[i];
				nListInvD[i] = listInvD[i];
			}
		}

		BDELETE(listX);
		BDELETE(listY);
		BDELETE(listInvD);
		listX = nListX;
		listY = nListY;
		listInvD = nListInvD;
	}

	// now construct linear Bzier curves
	for(i=0; i<len-1; i++)
	{
		double a = listX[i];
		double b = listX[i+1];
		double pX[] = {a, (2.0/3.0)*a + (1.0/3.0)*b, (1.0/3.0)*a + (2.0/3.0)*b, b};
		double c = listY[i];
		double d = listY[i+1];
		double pY[] = {c, (2.0/3.0)*c + (1.0/3.0)*d, (1.0/3.0)*c + (2.0/3.0)*d, d};
		double e = listInvD[i];
		double f = listInvD[i+1];
		double pInvD[] = {e, (2.0/3.0)*e + (1.0/3.0)*f, (1.0/3.0)*e + (2.0/3.0)*f, f};
		bezierVector->push_back(NEW3(Bezier, pX, pY, pInvD));
	}

	BDELETE(listX);
	BDELETE(listY);
	BDELETE(listInvD);
}

// precision = quadratic distance
void BezierPath::Fit(double precision)
{
	// accumulated quadratic error
	std::vector<double> *accErr = NEW(std::vector<double>);
	int i;
	for(i=0; i<GetLength(); i++)
	{
		accErr->push_back(0.0);
	}

	bool haveJoined;
	do
	{
		#ifndef NDEBUG
		double oldStartX = GetX(0, 0);
		double oldStartY = GetY(0, 0);
		double oldEndX = GetX(bezierVector->size()-1, 3);
		double oldEndY = GetY(bezierVector->size()-1, 3);
		#endif

		// Did we join curves this time?
		haveJoined = false;
		// joined curves go here:
		std::vector<Bezier*> *nb = NEW(std::vector<Bezier*>);
		// ... and their accumulated quadratic errors here:
		std::vector<double> *nAccErr = NEW(std::vector<double>);

		bool haveJoinedLastCurve = false;
		int i;
		for(i=0; i<GetLength()-1; i++) // not executed if getLength() == 1
		{
			double minDistance = 1e100;
			Bezier *bez;
			Bezier *minBez = NEW(Bezier);
			double t; // parameter value where to inversely subdivide

			// possible improvement: replace this brute force testing by something intelligent
			for(t=0.02; t<1.0; t+=0.02)
			{
				bez = NEW(Bezier);
				double distance = bez->FitJoin(*(*bezierVector)[i], *(*bezierVector)[i+1], t);
				if (distance < minDistance)
				{
					minDistance = distance;
					DELETE(minBez);
					minBez = bez;
				}
				else
				{
					DELETE(bez);
					#ifndef NDEBUG
					bez = 0;
					#endif
				}
			}

			// largest of the accumulated errors of the two to-be-joined curves
			double maxOrigAccErr = ((*accErr)[i+1] > (*accErr)[i]) ? (*accErr)[i+1] : (*accErr)[i];

			// Is the error small enough?
			if(minDistance + maxOrigAccErr < precision)
			{
				nb->push_back(minBez);
				nAccErr->push_back(minDistance + maxOrigAccErr);
				if(i == GetLength()-2)
				{
					haveJoinedLastCurve = true;
				}
				i++; // go on with curve i+2
				haveJoined = true;
			}
			else
			{
				delete minBez;
				#ifndef NDEBUG
				minBez = 0;
				#endif
				// Caution: need a copy, in case (*bezierVector)[i] is deleted
				nb->push_back( NEW1(Bezier, *(*bezierVector)[i]) );
				nAccErr->push_back((*accErr)[i]);
			}
		}

		if(haveJoined)
		{
			if(!haveJoinedLastCurve) // if the last one remains (e.g. 6 curves -> 2+1+2+2+1)
			{
				// Caution: need a copy, in case (*bezierVector)[...] is deleted
				nb->push_back( NEW1(Bezier, *(*bezierVector)[GetLength()-1]) );
				nAccErr->push_back((*accErr)[GetLength()-1]);
			}

			// delete old curves
			int j;
			for(j=0; j<bezierVector->size(); j++)
			{
				DELETE((*bezierVector)[j]);
				#ifndef NDEBUG
				(*bezierVector)[j] = 0;
				#endif
			}
			DELETE(bezierVector);
			bezierVector = nb;
			DELETE(accErr);
			accErr = nAccErr;
		}
		else
		{
			// delete new curves
			int j;
			for(j=0; j<nb->size(); j++)
			{
				DELETE((*nb)[j]);
				#ifndef NDEBUG
				(*nb)[j] = 0;
				#endif
			}
			DELETE(nb);
			DELETE(nAccErr);
			#ifndef NDEBUG
			nb = 0;
			nAccErr = 0;
			#endif
		}

		#ifndef NDEBUG
		assert(oldStartX == GetX(0, 0));
		assert(oldStartY == GetY(0, 0));
		assert(oldEndX == GetX(bezierVector->size()-1, 3));
		assert(oldEndY == GetY(bezierVector->size()-1, 3));
		if(bezierVector->size() > 1)
		{
			int j;
			for(j=0; j<bezierVector->size()-1; j++)
			{
				assert( GetX(j, 3) == GetX(j+1, 0) );
				assert( GetY(j, 3) == GetY(j+1, 0) );
			}
		}
		#endif

	} while(haveJoined);

	delete accErr;
}

double BezierPath::GetX(int i, int num) const
{
	assert((num>=0)&&(num<=3));
	assert((i>=0)&&(i<GetLength()));
	return (*bezierVector)[i]->GetX(num);
}

double BezierPath::GetY(int i, int num) const
{
	assert((num>=0)&&(num<=3));
	assert((i>=0)&&(i<GetLength()));
	return (*bezierVector)[i]->GetY(num);
}

double BezierPath::GetInvD(int i, int num) const
{
	assert((num>=0)&&(num<=3));
	assert((i>=0)&&(i<GetLength()));
	return (*bezierVector)[i]->GetInvD(num);
}

int BezierPath::GetLength() const
{
	return bezierVector->size();
}

double BezierPath::GetXt(int i, double t) const
{
	assert((i>=0)&&(i<GetLength()));
	return (*bezierVector)[i]->GetXt(t);
}

double BezierPath::GetYt(int i, double t) const
{
	assert((i>=0)&&(i<GetLength()));
	return (*bezierVector)[i]->GetYt(t);
}

double BezierPath::GetInvDt(int i, double t) const
{
	assert((i>=0)&&(i<GetLength()));
	return (*bezierVector)[i]->GetInvDt(t);
}

double BezierPath::GetDerivt(int i, double t) const
{
	assert((i>=0)&&(i<GetLength()));
	return (*bezierVector)[i]->GetDerivt(t);
}