// (c) 2002 by Jrn Loviscach

#include <assert.h>
#include <math.h>
#include <fstream>
#include <stdexcept>

#include "Bitmap.h"
#include "adapt.h"

Bitmap::Bitmap()
: hasBMP(false), imageData(0)
{}

Bitmap::Bitmap(const Bitmap& in)
{
	assert(in.hasBMP);

	width_ = in.width_;
	height_ = in.height_;
	rowLength = in.rowLength;

	imageData = BNEW(unsigned char, rowLength*height_);
	memcpy(imageData, in.imageData, rowLength*height_); 

	hasBMP = true;
}

Bitmap::~Bitmap()
{
	if(hasBMP)
	{
		hasBMP = false;
		BDELETE(imageData);
		imageData = 0;
	}
}

void Bitmap::CreateBlack (unsigned short width, unsigned short height)
{
	int width_1 = width;
	int height_1 = height;
	int rowLength_1 = 3*width_1;
	if(rowLength_1 % 4 != 0)
	{
		rowLength_1 = 4*(1+rowLength_1/4); // next multiple of 4
	}

	unsigned char* imageData_1 = BNEW(unsigned char, rowLength_1*height_1);
	memset(imageData_1, 0, rowLength_1*height_1);

	// for exception safety delete old bitmap only now
	if(hasBMP)
	{
		hasBMP = false;
		BDELETE(imageData);
		imageData = 0;
	}

	width_ = width_1;
	height_ = height_1;
	rowLength = rowLength_1;
	imageData = imageData_1;
	hasBMP = true;
}

void Bitmap::CreateWhite (unsigned short width, unsigned short height)
{
	int width_1 = width;
	int height_1 = height;
	int rowLength_1 = 3*width_1;
	if(rowLength_1 % 4 != 0)
	{
		rowLength_1 = 4*(1+rowLength_1/4); // next multiple of 4
	}

	unsigned char* imageData_1 = BNEW(unsigned char, rowLength_1*height_1);
	memset(imageData_1, 255, rowLength_1*height_1);

	// for exception safety delete old bitmap only now
	if(hasBMP)
	{
		hasBMP = false;
		BDELETE(imageData);
		imageData = 0;
	}

	width_ = width_1;
	height_ = height_1;
	rowLength = rowLength_1;
	imageData = imageData_1;
	hasBMP = true;
}

void Bitmap::White()
{
	assert(hasBMP);
	memset(imageData, 255, rowLength*height_);
}

// Copy bitmap and add a pixel wide frame to facilitate
// operations with 3x3 matrices
void Bitmap::Frame(const Bitmap& in)
{
	int width_1 = in.GetWidth()+2;
	int height_1 = in.GetHeight()+2;
	int rowLength_1 = 3*width_1;
	if(rowLength_1 % 4 != 0)
	{
		rowLength_1 = 4*(1+rowLength_1/4); // next multiple of 4
	}

	unsigned char* imageData_1 = BNEW(unsigned char, rowLength_1*height_1);

	for(int x=0; x<in.GetWidth(); ++x)
	{
		for(int y=0; y<in.GetHeight(); ++y)
		{
			for(int k=0; k<3; ++k)
			{
				unsigned char v = in.GetPixel(x, y, k);
				SetPixel(x+1, y+1, k, v);
			}
		}
	}
	
	// fill margin rows and columns by cloning
	for(int x=0; x<in.GetWidth(); ++x)
	{
		for(int k=0; k<3; ++k)
		{
			unsigned char v = in.GetPixel(x, 0, k);
			SetPixel(x+1, 0, k, v);

			v = in.GetPixel(x, in.GetHeight()-1, k);
			SetPixel(x+1, in.GetHeight()+1, k, v);
		}
	}

	for(int y=0; y<in.GetHeight(); ++y)
	{
		for(int k=0; k<3; ++k)
		{
			unsigned char v = in.GetPixel(0, y, k);
			SetPixel(0, y+1, k, v);

			v = in.GetPixel(in.GetWidth()-1, y, k);
			SetPixel(in.GetWidth()+1, y+1, k, v);
		}
	}

	for(int k=0; k<3; ++k)
	{
		unsigned char v = in.GetPixel(0, 0, k);
		SetPixel(0, 0, k, v);

		v = in.GetPixel(in.GetWidth()-1, 0, k);
		SetPixel(in.GetWidth()+1, 0, k, v);

		v = in.GetPixel(0, in.GetHeight()-1, k);
		SetPixel(0, in.GetHeight()+1, k, v);

		v = in.GetPixel(in.GetWidth()-1, in.GetHeight()-1, k);
		SetPixel(in.GetWidth()+1, in.GetHeight()+1, k, v);
	}

	// for exception safety delete old bitmap only now
	if(hasBMP)
	{
		hasBMP = false;
		BDELETE(imageData);
		imageData = 0;
	}

	width_ = width_1;
	height_ = height_1;
	rowLength = rowLength_1;
	imageData = imageData_1;
	hasBMP = true;
}

// Extract silhouettes (red channel), creases (blue channel),
// and their inverse depth (blue channel) from the GBuffer.
// Needs a "framed" bitmap.
// angleSens measured in radians.
void Bitmap::Extract(const GBuffer& in, double depthSens, double angleSens, double minDepth, double maxDepth, bool intersctAsCreas)
{
	assert(hasBMP);

	double cosAngleSens = cos(angleSens);

	for(int x=1; x<in.GetWidth()-1; ++x)
	{
		for(int y=1; y<in.GetHeight()-1; ++y)
		{
			long objNum = in.GetObjNumber(x, y);
			if( objNum == -1 ) // i.e. background
			{
				continue;
			}

			// determine min and max depth on the 8-neighborhood
			double depthMin = in.GetDepth(x, y); // to start with
			double depthMax = depthMin;
			bool isNearBackground = false;
			bool drawIntersection = false;
			for(int x1=-1; x1<=1; ++x1)
			{
				for(int y1=-1; y1<=1; ++y1)
				{
					if( in.GetObjNumber(x+x1, y+y1) == -1 )
					{
						isNearBackground = true;
						continue; // can't take depth, because it's undefined here
					}
					if( in.GetObjNumber(x+x1, y+y1) != objNum )
					{
						// draw boundary line only inside the object with lower number
						// to avoid double lines
						if( in.GetObjNumber(x+x1, y+y1) > objNum )
						{
							drawIntersection = true;
						}
					}

					double v = in.GetDepth(x+x1, y+y1);
					if(v > depthMax)
					{
						depthMax = v;
					}
					else if(v < depthMin)
					{
						depthMin = v;
					}
				}
			}	

			// store inverse depth in blue channel
			double invD = 255.0*(maxDepth-depthMin)/(maxDepth-minDepth);
			if( invD >= 255.5 )
			{
				invD = 255.5;
			}
			else if( invD < 0.5 )
			{
				invD = 0.5;
			}			
			SetPixel(x, y, 2, (int)invD);

			// mark only front pixels (i.e. inside a silhouette)
			if( ( in.GetDepth(x, y) < depthMax - depthSens )
				|| isNearBackground
				|| ( drawIntersection && !intersctAsCreas) )
			{
				// silhouette to red
				SetPixel(x, y, 0, 255);
			}
			else if( depthMax - depthMin <= depthSens )
			{
				// no strong depth variations here

				if( drawIntersection && intersctAsCreas )
				{
					SetPixel(x, y, 1, 255);
				}
				else
				{
					// test for crease
					double minScalarProd = 1.0;
					for(int x2=0; x2<=1; ++x2) // only look right and above, to thin lines
					{
						for(int y2=1-x2; y2<=1; ++y2) // if x2 == 0 skip y2 == 0 
						{
							double scalarProd =
								in.GetNormalX(x-x2, y-y2)*in.GetNormalX(x+x2, y+y2)
								+ in.GetNormalY(x-x2, y-y2)*in.GetNormalY(x+x2, y+y2)
								+ in.GetNormalZ(x-x2, y-y2)*in.GetNormalZ(x+x2, y+y2);
							if(scalarProd < minScalarProd)
							{
								minScalarProd = scalarProd;
							}
						}
					}
					if(minScalarProd < cosAngleSens)
					{
						// crease to green
						SetPixel(x, y, 1, 255);
					}
				}
			}
		}
	}
}

// erase crease candidates near silhouettes
void Bitmap::Clean(int gapAroundSilh)
{
	assert(hasBMP);

	for(int x=0; x<GetWidth(); ++x)
	{
		for(int y=0; y<GetHeight(); ++y)
		{
			if(GetPixel(x, y, 1) == 255) // marked as crease?
			{
				bool isNearContour = false;
				int x1, y1;
				for(x1=-gapAroundSilh; x1<=gapAroundSilh; ++x1)
				{
					for(y1=-gapAroundSilh; y1<=gapAroundSilh; ++y1)
					{
						if(x1*x1 + y1*y1 > gapAroundSilh*gapAroundSilh) // outside circle
						{
							continue;
						}

						int x2 = x+x1;
						x2 = (x2<0) ? 0 : x2;
						x2 = (x2>=GetWidth()) ? GetWidth()-1 : x2;

						int y2 = y+y1;
						y2 = (y2<0) ? 0 : y2;
						y2 = (y2>=GetHeight()) ? GetHeight()-1 : y2;

						if(GetPixel(x2, y2, 0) == 255)
						{
							isNearContour = true;
						}
					}
				}
				if(isNearContour)
				{
					SetPixel(x, y, 1, 0);
				}
			}
		}
	}
}

// needs a "framed" picture
void Bitmap::Thin()
{
	assert(hasBMP);

	Bitmap* p1 = this;
	Bitmap* p2 = (Bitmap*)NEW(Bitmap);
	p2->CreateBlack(p1->GetWidth(), p1->GetHeight());
	Bitmap* aux; // to aid in swapping
	bool someCleared;

	// thin silhouettes
	do
	{
		someCleared = p1->thinOnce(p2, 0);
		aux = p2; // role change
		p2 = p1;
		p1 = aux;
	} while(someCleared);

	// thin creases
	do
	{
		someCleared = p1->thinOnce(p2, 1);
		aux = p2; // role change
		p2 = p1;
		p1 = aux;
	} while(someCleared);

	if(p1 != this) // data is in p2 instead of this
	{
		for (unsigned long i=0; i<rowLength*height_; ++i)
		{
			imageData[i] = p1->imageData[i];
		}
		DELETE(p1);
	}
	else
	{
		DELETE(p2);
	}
}

// see W.K. Pratt, Digital image processing (2nd ed), John Wiley and Sons, 1991
// Returns true, if at least one pixel has been cleared, else false.
// needs a "framed" picture
bool Bitmap::thinOnce(Bitmap* out, int channel) const
{
	assert(out!=0);
	assert(channel == 0 || channel == 1);

	out->CreateBlack(GetWidth(), GetHeight());

	bool* markedPixels = BNEW(bool, GetWidth() * GetHeight());

	bool somePixelDeleted = false;

	// relative coordinates of neighbor pixels
	const int xr[] = {1, 1, 1, 0, -1, -1, -1, 0, 0};
	const int yr[] = {1, 0, -1, -1, -1, 0, 1, 1, 0};

	const int thin1[] = { 11, 15, 27, 29, 31, 41, 45, 57, 61, 63,
						  109, 113, 121, 125, 127, 131, 135, 161,
						  225, 241, 249, 253, 263, 267, 271, 283,
						  287, 319, 383, 387, 391, 399, 415, 449,
						  451, 455, 463, 479, 481, 483, 487, 497,
						  499, 503, 505, 509 };

	const int thin2[] = { 3, 5, 7, 13, 17, 25, 27, 43, 45, 47, 49,
						  57, 59, 61, 63, 71, 75, 77, 79, 83, 85,
						  87, 91, 93, 95, 97, 101, 107, 115, 117,
						  119, 121, 123, 125, 127, 129, 135, 139,
						  143, 147, 149, 151, 155, 157, 159, 163,
						  165, 167, 169, 173, 179, 181, 183, 185,
						  187, 189, 191, 193, 197, 203, 211, 213,
						  215, 219, 221, 223, 227, 229, 233, 243,
						  245, 247, 249, 251, 253, 255, 259, 267,
						  275, 277, 279, 281, 285, 287, 293, 295,
						  297, 299, 301, 303, 305, 309, 311, 313,
						  315, 317, 319, 325, 327, 329, 331, 333,
						  335, 337, 339, 341, 343, 345, 347, 349,
						  351, 357, 359, 361, 363, 365, 367, 369,
						  371, 373, 375, 377, 379, 381, 383, 385,
						  395, 401, 403, 405, 407, 413, 415, 419,
						  421, 423, 425, 429, 431, 437, 439, 441,
						  443, 445, 447, 453, 455, 457, 459, 461,
						  463, 465, 467, 469, 471, 473, 475, 477,
						  479, 485, 487, 489, 491, 493, 495, 497,
						  499, 501, 503, 505, 507, 509, 511 };

	bool thin1LUT[512];
	for(int i=0; i<512; ++i)
	{
		thin1LUT[i] = false;
	}
	for(int i=0; i<sizeof(thin1)/sizeof(int); ++i)
	{
		thin1LUT[thin1[i]] = true;
	}

	bool thin2LUT[512];
	for(int i=0; i<512; i++)
	{
		thin2LUT[i] = false;
	}
	for(int i=0; i<sizeof(thin2)/sizeof(int); ++i)
	{
		thin2LUT[thin2[i]] = true;
	}

	// mark candidates
	for(int x=1; x<GetWidth()-1; ++x)
	{
		for(int y=1; y<GetHeight()-1; ++y)
		{
			int stackedNeighbors = 0;					
			for(int i=0; i<9; ++i)
			{
				stackedNeighbors = (stackedNeighbors<<1) | ( (GetPixel(x+xr[i], y+yr[i], channel) == 255)?1:0 );	
			}

			markedPixels[x+y*GetWidth()] = thin1LUT[stackedNeighbors];
		}
	}
	
	// final choice
	for(int x=1; x<GetWidth()-1; ++x)
	{
		for(int y=1; y<GetHeight()-1; ++y)
		{
			// copy pixel
			out->SetPixel(x, y, 0, GetPixel(x, y, 0));
			out->SetPixel(x, y, 1, GetPixel(x, y, 1));
			out->SetPixel(x, y, 2, GetPixel(x, y, 2));

			// need to delete?
			if( GetPixel(x, y, channel) == 255 )
			{
				if(!markedPixels[x + y*GetWidth()])
				{
					out->SetPixel(x, y, channel, 255);
					out->SetPixel(x, y, 2, GetPixel(x, y, 2));
				}
				else
				{
					int stackedNeighbors = 0;					
					for(int i=0; i<9; ++i)
					{
						stackedNeighbors = (stackedNeighbors<<1) | ( (markedPixels[x+xr[i] + (y+yr[i])*GetWidth()])?1:0 );	
					}

					if( ! thin2LUT[stackedNeighbors] ) // then delete
					{
						out->SetPixel(x, y, channel, 0);
						somePixelDeleted = true;
					}
				}
			}
		}
	}

	BDELETE(markedPixels);

	return somePixelDeleted;
}

// No gamma neeeded here!
void Bitmap::Blend(const Bitmap& bm)
{
	for(int x=0; x<GetWidth(); ++x)
	{
		for(int y=0; y<GetHeight(); ++y)
		{
			for(int c=0; c<3; ++c)
			{
				const int a = 255;
				const int b = a/2;
				SetPixel(x, y, c, (unsigned char)((GetPixel(x, y, c)*bm.GetPixel(x, y, c) + b)/a));
			}
		}
	}
}

void Bitmap::InitFromRGBA(unsigned char* RGBA, int width, int height)
{
	int width_1 = width;
	int height_1 = height;
	int rowLength_1 = 3*width_1;
	if(rowLength_1 % 4 != 0)
	{
		rowLength_1 = 4*(1+rowLength_1/4); // next multiple of 4
	}

	unsigned char* imageData_1 =
		BNEW(unsigned char, rowLength_1*height_1);

	for(int y=0; y<height_1; ++y)
	{
		for(int x=0; x<width_1; ++x)
		{
			for(int i=0; i<3; ++i)
			{
				imageData_1[3*x+rowLength_1*y+i] = RGBA[4*(x+width_1*y)+i];
			}
		}
	}

	// for exception safety delete old bitmap only now
	if(hasBMP)
	{
		hasBMP = false;
		BDELETE(imageData);
		imageData = 0;
	}

	width_ = width_1;
	height_ = height_1;
	rowLength = rowLength_1;
	imageData = imageData_1;
	hasBMP = true;
}

void Bitmap::StoreToRGBA(unsigned char* RGBA)
{
	assert(hasBMP);

	for(int y=0; y<height_; ++y)
	{
		for(int x=0; x<width_; ++x)
		{
			for(int i=0; i<3; ++i)
			{
				RGBA[4*(x+width_*y)+i] = imageData[3*x+rowLength*y+i];
			}
		}
	}
}