// (c) 2002 by Jrn Loviscach

#include <assert.h>
#include <math.h>

#include "adapt.h"
#include "brush.h"
#include "bitmap.h"
#include "bezierpath.h"
#include "bezierpathcollection.h"

// first call initializes (because of "static")
Brush* GetBrush(int i)
{
	Brush* b = 0;
	static bool isOK = true;

	if(!isOK)
	{
		return 0;
	}

	try
	{
		static Brush* brushColl[numberOfBrushTypes] =
		{
			NEW(Pen),
			NEW(InkBrush),
			NEW(WiggleLines),
			NEW(Pencil)
		};

		b = brushColl[i];
	}
	catch(exception&)
	{
		isOK = false;
		return 0;
	}

	return b;
}

Brush::~Brush()
{
	for(int i=0; i<GetNumParams(); ++i)
	{
		BDELETE(myParams[i].name);
	}
	BDELETE(myParams);
}

int Brush::GetNumParams() const
{
	return numParams;
}

std::string Brush::GetParamName(int i) const
{
	assert(i>=0 || i<GetNumParams());
	return *myParams[i].name;
}

bool Brush::IsIntParam(int i) const
{
	assert(i>=0 || i<GetNumParams());
	return myParams[i].isInt;
}

double Brush::MinParamVal(int i) const
{
	assert(i>=0 || i<GetNumParams());
	return myParams[i].minVal;
}

double Brush::MaxParamVal(int i) const
{
	assert(i>=0 || i<GetNumParams());
	return myParams[i].maxVal;
}

double Brush::StdParamVal(int i) const
{
	assert(i>=0 || i<GetNumParams());
	return myParams[i].stdVal;
}

// caller has to dispose
Brush::Param* Brush::GetCopyOfStdParams() const
{
	Param* copy = BNEW(Param, GetNumParams());
	for(int i=0; i<GetNumParams(); ++i)
	{
		if(IsIntParam(i))
		{
			copy[i].i_ = (int)(StdParamVal(i) + 0.5);
		}
		else
		{
			copy[i].f_ = StdParamVal(i);
		}
	}
	return copy;
}

// caller has to dispose
double* Brush::ConvParamsIFtoD(const Param* params) const
{
	double* paramsD = BNEW(double, GetNumParams());
	for(int i=0; i<GetNumParams(); ++i)
	{
		if(IsIntParam(i))
		{
			paramsD[i] = params[i].i_;
		}
		else
		{
			paramsD[i] = params[i].f_;
		}
	}
	return paramsD;
}

// needed to avoid range overflow when animating
void Brush::clampParams(double* params) const
{
	for(int i=0; i<GetNumParams(); ++i)
	{
		if(IsIntParam(i))
		{
			params[i]+=0.5; // so that we simply can cast to (int) later on with no bad roundoff errors
			if(params[i]<MinParamVal(i) + 0.1)
			{
				params[i] = MinParamVal(i) + 0.1;
			}
			else if(params[i]>MaxParamVal(i) + 0.1)
			{
				params[i] = MaxParamVal(i) + 0.1;
			}
		}
		else // param i has float value
		{
			if(params[i]<MinParamVal(i))
			{
				params[i] = MinParamVal(i);
			}
			else if(params[i]>MaxParamVal(i))
			{
				params[i] = MaxParamVal(i);
			}
		}
	}
}

//////////////////////////////////////////////////////////////////

Pen::Pen()
{
	numParams = 3;
	const ParamData dummy[] = {
		{NEW1(std::string, "Supersampling"), true, 1.0, 4.0, 2.0},
		{NEW1(std::string, "Radius"), false, 0.0, 100.0, 0.7},
		{NEW1(std::string, "Depth->Radius"), true, 0.0, 100.0, 50.0}
	};
	myParams = BNEW(ParamData, numParams);
	memcpy(myParams, dummy, numParams*sizeof(ParamData));
}

std::string Pen::GetName() const
{
	return "Pen";
}

void Pen::Paint(
		const BezierPathCollection& bezColl,
		Bitmap& bm,
		const double* params,
		double time,
		double targetDpi,
		double gamma) const
{
	double* clmpParams = BNEW(double, GetNumParams());
	memcpy(clmpParams, params, GetNumParams()*sizeof(double));
	clampParams(clmpParams);

	int superSamp = (int)clmpParams[0];
	double superRadius = superSamp*targetDpi/72.0*clmpParams[1];

	int superWidth = superSamp*bm.GetWidth();
	int superHeight = superSamp*bm.GetHeight();
	bool* superImage = BNEW(bool, superWidth*superHeight);
	for(int i=0; i<superWidth*superHeight; ++i)
	{
		superImage[i] = false;
	}

	double widthFactor = superWidth/bezColl.GetWidth();
	double heightFactor = superHeight/bezColl.GetHeight();

	for(int i=0; i<bezColl.GetLength(); ++i)
	{
		const BezierPath* bp = bezColl.GetPath(i);

		for(int j=0; j<bp->GetLength(); ++j)
		{
			double t=0.0;
			do {
				double xc = widthFactor*bp->GetXt(j, t);
				double yc = heightFactor*bp->GetYt(j, t);
				double depthToRadius = 1.0+(-0.9+1.8*bp->GetInvDt(j, t)/255.0)*clmpParams[2]/100.0;
				int radius = (int)(superRadius*depthToRadius+0.5);
				for(int xd=-radius; xd<=radius; ++xd)
				{
					for(int yd=-radius; yd<=radius; ++yd)
					{
						int r2 = xd*xd+yd*yd;
						if(r2 > radius*radius) continue;

						int x = xc+xd;
						if(x<0) continue;
						if(x>=superWidth) continue;

						int y = yc+yd;
						if(y<0) continue;
						if(y>=superHeight) continue;

						superImage[y*superWidth+x]=true;
					}
				}

				// space evenly along curve
				double d = bp->GetDerivt(j, t);
				t+=2.0/(widthFactor+heightFactor)/(0.01+d);
			} while (t<1.000001);
		}
	}

	// supersampling
	for(int x=0; x<bm.GetWidth(); ++x)
	{
		for(int y=0; y<bm.GetHeight(); ++y)
		{
			int n = 0; // becomes number of set pixels
			for(int xd=0; xd<superSamp; ++xd)
			{
				for(int yd=0; yd<superSamp; ++yd)
				{
					if(superImage[(y*superSamp+yd)*superWidth + x*superSamp+xd])
					{
						++n;
					}
				}
			}
			// 1.0 - ...: paint black, not white
			double val = pow(1.0-n/(double)(superSamp*superSamp), 1.0/gamma);
			// no gamma needed for blending with former value because we are multiplying
			int intVal = (int)(0.5+255.0*val);
			for(int c=0; c<3; ++c)
			{
				const int a = 255;
				const int b = a/2;
				bm.SetPixel(x, y, c, (unsigned char)((intVal*bm.GetPixel(x, y, c) + b)/a));
			}
		}
	}

	BDELETE(superImage);
	BDELETE(clmpParams);
}

//////////////////////////////////////////////////////////////////

InkBrush::InkBrush()
{
	numParams = 8;
	const ParamData dummy[] = {
		{NEW1(std::string, "Supersampling"), true, 1.0, 4.0, 2.0},
		{NEW1(std::string, "Number of Hairs"), true, 1.0, 20.0, 3.0},
		{NEW1(std::string, "Distance of Hairs"), false, 0.0, 100.0, 1.0},
		{NEW1(std::string, "Radius of Hairs"), false, 0.0, 100.0, 0.7},
		{NEW1(std::string, "Irregularity of Distance"), true, 0.0, 100.0, 10.0},
		{NEW1(std::string, "Irregularity of Radius"), true, 0.0, 100.0, 10.0},
		{NEW1(std::string, "Depth->Hair Distance"), true, 0.0, 100.0, 50.0},
		{NEW1(std::string, "Fade Out"), true, 0.0, 100.0, 50.0}
	};
	myParams = BNEW(ParamData, numParams);
	memcpy(myParams, dummy, numParams*sizeof(ParamData));
}

std::string InkBrush::GetName() const
{
	return "Ink Brush";
}

void InkBrush::Paint(
		const BezierPathCollection& bezColl,
		Bitmap& bm,
		const double* params,
		double time,
		double targetDpi,
		double gamma) const
{
	double* clmpParams = BNEW(double, GetNumParams());
	memcpy(clmpParams, params, GetNumParams()*sizeof(double));
	clampParams(clmpParams);

	int superSamp = (int)clmpParams[0];
	int numHairs = (int)clmpParams[1];
	double hairDist = clmpParams[2];
	double hairSuperRadius = superSamp*targetDpi/72.0*clmpParams[3];

	int superWidth = superSamp*bm.GetWidth();
	int superHeight = superSamp*bm.GetHeight();
	bool* superImage = BNEW(bool, superWidth*superHeight);
	for(int i=0; i<superWidth*superHeight; ++i)
	{
		superImage[i] = false;
	}

	double widthFactor = superWidth/bezColl.GetWidth();
	double heightFactor = superHeight/bezColl.GetHeight();

	double* xDelta = BNEW(double, numHairs);
	double* yDelta = BNEW(double, numHairs);
	double* radius = BNEW(double, numHairs);

	for(int i=0; i<numHairs; ++i)
	{
		double random1 = 1000.0*sqrt(i*2.652474566);
		random1 += -floor(random1)-0.5; // centered around 0.0
		double irregDist = 1.0+((int)clmpParams[4])/100.0*1.9*random1;

		double random2 = 10000.0*sqrt(i*3.374567986);
		random2 += -floor(random2)-0.5; // centered around 0.0
		double irregRad = 1.0+((int)clmpParams[5])/100.0*1.9*random2;

		xDelta[i] = superSamp*targetDpi/72.0*hairDist*irregDist*cos((i*2*3.1415)/numHairs);
		yDelta[i] = superSamp*targetDpi/72.0*hairDist*irregDist*sin((i*2*3.1415)/numHairs);
		radius[i] = hairSuperRadius*irregRad;
	}

	for(int i=0; i<bezColl.GetLength(); ++i)
	{
		const BezierPath* bp = bezColl.GetPath(i);

		// calculate curve length
		double curveLen = 0.0;
		for(int j=0; j<bp->GetLength(); ++j)
		{
			double t=0.0;
			do {
				double d = bp->GetDerivt(j, t);
				curveLen += d;
				// space evenly along curve, same as below
				t+=2.0/(widthFactor+heightFactor)/(0.01+d);
			} while (t<1.000001);
		}

		// now paint
		double lenBehind = 0.0; // How far are we?
		for(int j=0; j<bp->GetLength(); ++j)
		{
			double t=0.0;
			do {
				double xc = widthFactor*bp->GetXt(j, t);
				double yc = heightFactor*bp->GetYt(j, t);
				double depthToDist = 1.0+(-0.9+1.8*bp->GetInvDt(j, t)/255.0)*clmpParams[6]/100.0;
				double fadeOutFactor = 1.0-lenBehind/curveLen*clmpParams[7]/100.0;
				for(int k=0; k<numHairs; ++k)
				{
					int rMax = (int)ceil(radius[k]*fadeOutFactor);
					int xcc = (int)(0.5+xc+xDelta[k]*depthToDist);
					int ycc = (int)(0.5+yc+yDelta[k]*depthToDist);
					for(int xd=-rMax; xd<=rMax; ++xd)
					{
						for(int yd=-rMax; yd<=rMax; ++yd)
						{
							int r = sqrt(xd*xd+yd*yd);
							if(r > radius[k]*fadeOutFactor) continue;

							int x = xcc+xd;
							if(x<0) continue;
							if(x>=superWidth) continue;

							int y = ycc+yd;
							if(y<0) continue;
							if(y>=superHeight) continue;

							superImage[y*superWidth+x]=true;
						}
					}
				}
				double d = bp->GetDerivt(j, t);
				lenBehind += d;
				// space evenly along curve, see above
				t+=2.0/(widthFactor+heightFactor)/(0.01+d);
			} while (t<1.000001);
		}
	}

	BDELETE(xDelta);
	BDELETE(yDelta);
	BDELETE(radius);

	// supersampling
	for(int x=0; x<bm.GetWidth(); ++x)
	{
		for(int y=0; y<bm.GetHeight(); ++y)
		{
			int n = 0; // becomes number of set pixels
			for(int xd=0; xd<superSamp; ++xd)
			{
				for(int yd=0; yd<superSamp; ++yd)
				{
					if(superImage[(y*superSamp+yd)*superWidth + x*superSamp+xd])
					{
						++n;
					}
				}
			}
			// 1.0 - ...: paint black, not white
			double val = pow(1.0-n/(double)(superSamp*superSamp), 1.0/gamma);
			// no gamma needed for blending with former value because we are multiplying
			int intVal = (int)(0.5+255.0*val);
			for(int c=0; c<3; ++c)
			{
				const int a = 255;
				const int b = a/2;
				bm.SetPixel(x, y, c, (unsigned char)((intVal*bm.GetPixel(x, y, c) + b)/a));
			}
		}
	}

	BDELETE(superImage);
	BDELETE(clmpParams);
}

//////////////////////////////////////////////////////////////////

WiggleLines::WiggleLines()
{
	numParams = 7;
	const ParamData dummy[] = {
		{NEW1(std::string, "Supersampling"), true, 1.0, 4.0, 2.0},
		{NEW1(std::string, "Number of Lines"), true, 1.0, 20.0, 3.0},
		{NEW1(std::string, "Radius"), false, 0.0, 100.0, 0.7},
		{NEW1(std::string, "Depth->Radius"), true, 0.0, 100.0, 50.0},
		{NEW1(std::string, "Amount"), false, 0.0, 1000.0, 50.0},
		{NEW1(std::string, "Wavelength"), false, 0.1, 1000.0, 50.0},
		{NEW1(std::string, "Frame-to-Frame Variation"), true, 0.0, 100.0, 50.0}
	};
	myParams = BNEW(ParamData, numParams);
	memcpy(myParams, dummy, numParams*sizeof(ParamData));
}

std::string WiggleLines::GetName() const
{
	return "Wiggle Lines";
}

void WiggleLines::Paint(
		const BezierPathCollection& bezColl,
		Bitmap& bm,
		const double* params,
		double time,
		double targetDpi,
		double gamma) const
{
	double* clmpParams = BNEW(double, GetNumParams());
	memcpy(clmpParams, params, GetNumParams()*sizeof(double));
	clampParams(clmpParams);

	int superSamp = (int)clmpParams[0];
	double superRadius = superSamp*targetDpi/72.0*clmpParams[2];

	int superWidth = superSamp*bm.GetWidth();
	int superHeight = superSamp*bm.GetHeight();
	bool* superImage = BNEW(bool, superWidth*superHeight);
	for(int i=0; i<superWidth*superHeight; ++i)
	{
		superImage[i] = false;
	}

	double widthFactor = superWidth/bezColl.GetWidth();
	double heightFactor = superHeight/bezColl.GetHeight();

	int numberOfLines = (int)clmpParams[1];
	double amountSuper = superSamp*clmpParams[4]*targetDpi/72.0;
	double wavelengthSuper = superSamp*clmpParams[5]*targetDpi/72.0;
	double variation = ((int)clmpParams[6])/100.0;

	for(int i=0; i<bezColl.GetLength(); ++i)
	{
		const BezierPath* bp = bezColl.GetPath(i);

		for(int j=0; j<bp->GetLength(); ++j)
		{
			double t=0.0;
			do {
				double xc = widthFactor*bp->GetXt(j, t);
				double yc = heightFactor*bp->GetYt(j, t);
				double depthToRadius = 1.0+(-0.9+1.8*bp->GetInvDt(j, t)/255.0)*clmpParams[3]/100.0;

				for(int i=0; i<numberOfLines; ++i)
				{
					// different pseudo-random number for each of the lines
					double iVar = 1658.576*sin(0.614673*i);
					iVar -= floor(iVar);
					// more or less different pseudo-random number for each frame
					double tVar = (1.0 + variation*time*0.000934751);

					double tx = xc/wavelengthSuper;
					double ty = yc/wavelengthSuper;
					double fx = tx - floor(tx);
					double fy = ty - floor(ty);
					double sx[3]; // weights of quadratic B-Spline
					sx[0] = (1.0-fx)*(1.0-fx)/2.0;
					sx[1] = fx*(1.0-fx)+0.5;
					sx[2] = fx*fx/2.0;
					double sy[3]; // weights of quadratic B-Spline
					sy[0] = (1.0-fy)*(1.0-fy)/2.0;
					sy[1] = fy*(1.0-fy)+0.5;
					sy[2] = fy*fy/2.0;

					double xOff = 0.0;
					double yOff = 0.0;
					for(int xi=0; xi<3; ++xi)
					{
						for(int yi=0; yi<3; ++yi)
						{
							double xr = tVar*7656.567*sin(5343.937+iVar+13.523*sqrt(floor(tx)+xi)+floor(ty)+yi);
							xOff += sx[xi]*sy[yi]*(xr - floor(xr));
							double yr = tVar*4875.265*sin(7823.251+iVar+15.671*sqrt(floor(tx)+xi)+floor(ty)+yi);
							yOff += sx[xi]*sy[yi]*(yr - floor(yr));
						}
					}

					int xe = xc + amountSuper*(xOff-0.5);
					int ye = yc + amountSuper*(yOff-0.5);

					int radius = (int)(superRadius*depthToRadius+0.5);
					for(int xd=-radius; xd<=radius; ++xd)
					{
						for(int yd=-radius; yd<=radius; ++yd)
						{
							int r2 = xd*xd+yd*yd;
							if(r2 > radius*radius) continue;
							
							int x = xe+xd;
							if(x<0) continue;
							if(x>=superWidth) continue;

							int y = ye+yd;
							if(y<0) continue;
							if(y>=superHeight) continue;

							superImage[y*superWidth+x]=true;
						}
					}
				}
				// space evenly along curve
				double d = bp->GetDerivt(j, t);
				t+=2.0/(widthFactor+heightFactor)/(0.01+d)/sqrt(1.0+4.0*(amountSuper/wavelengthSuper)*(amountSuper/wavelengthSuper));
			} while (t<1.000001);
		}
	}

	// supersampling
	for(int x=0; x<bm.GetWidth(); ++x)
	{
		for(int y=0; y<bm.GetHeight(); ++y)
		{
			int n = 0; // becomes number of set pixels
			for(int xd=0; xd<superSamp; ++xd)
			{
				for(int yd=0; yd<superSamp; ++yd)
				{
					if(superImage[(y*superSamp+yd)*superWidth + x*superSamp+xd])
					{
						++n;
					}
				}
			}
			// 1.0 - ...: paint black, not white
			double val = pow(1.0-n/(double)(superSamp*superSamp), 1.0/gamma);
			// no gamma needed for blending with former value because we are multiplying
			int intVal = (int)(0.5+255.0*val);
			for(int c=0; c<3; ++c)
			{
				const int a = 255;
				const int b = a/2;
				bm.SetPixel(x, y, c, (unsigned char)((intVal*bm.GetPixel(x, y, c) + b)/a));
			}
		}
	}

	BDELETE(superImage);
	BDELETE(clmpParams);
}

//////////////////////////////////////////////////////////////////

Pencil::Pencil()
{
	numParams = 6;
	const ParamData dummy[] = {
		{NEW1(std::string, "Supersampling"), true, 1.0, 4.0, 1.0},
		{NEW1(std::string, "Radius"), false, 0.0, 100.0, 2.0},
		{NEW1(std::string, "Pressure"), true, 0.0, 100.0, 40.0},
		{NEW1(std::string, "Paper Structure Size"), false, 0.01, 50.0, 2.0},
		{NEW1(std::string, "Depth->Radius"), true, 0.0, 100.0, 50.0},
		{NEW1(std::string, "Depth->Pressure"), true, 0.0, 100.0, 20.0},
	};
	myParams = BNEW(ParamData, numParams);
	memcpy(myParams, dummy, numParams*sizeof(ParamData));
}

std::string Pencil::GetName() const
{
	return "Pencil";
}

void Pencil::Paint(
		const BezierPathCollection& bezColl,
		Bitmap& bm,
		const double* params,
		double time,
		double targetDpi,
		double gamma) const
{
	double* clmpParams = BNEW(double, GetNumParams());
	memcpy(clmpParams, params, GetNumParams()*sizeof(double));
	clampParams(clmpParams);

	int superSamp = (int)clmpParams[0];
	double superRadius = superSamp*targetDpi/72.0*clmpParams[1];

	int superWidth = superSamp*bm.GetWidth();
	int superHeight = superSamp*bm.GetHeight();
	unsigned char *depthSuperImage = new unsigned char[superWidth*superHeight];
	memset(depthSuperImage, 0, superWidth*superHeight);

	double widthFactor = superWidth/bezColl.GetWidth();
	double heightFactor = superHeight/bezColl.GetHeight();

	double pressure = ((int)clmpParams[2])/100.0;
	double structureSizeSuper = superSamp*clmpParams[3]*targetDpi/72.0;
	double depthToRadius = ((int)clmpParams[4])/100.0;
	double depthToPressure = ((int)clmpParams[5])/100.0;

	for(int i=0; i<bezColl.GetLength(); ++i)
	{
		const BezierPath* bp = bezColl.GetPath(i);

		for(int j=0; j<bp->GetLength(); ++j)
		{
			double t=0.0;
			do {
				double xc = widthFactor*bp->GetXt(j, t);
				double yc = heightFactor*bp->GetYt(j, t);
				double radFactor = 1.0+(-0.9+1.8*bp->GetInvDt(j, t)/255.0)*depthToRadius;
				int radius = (int)(superRadius*radFactor+0.5);
				int rad2 = radius*radius;

				double c = pressure+(bp->GetInvDt(j, t)/255.0 - 0.5)*depthToPressure;

				for(int xd=-radius; xd<=radius; ++xd)
				{
					for(int yd=-radius; yd<=radius; ++yd)
					{
						int r2 = xd*xd+yd*yd;
						if(r2 > rad2) continue;

						int x = xc+xd;
						if(x<0) continue;
						if(x>=superWidth) continue;

						int y = yc+yd;
						if(y<0) continue;
						if(y>=superHeight) continue;

						// multibly with edge decay
						double e = c*(1.0 - r2/(double)rad2); 
						if(e < 0.0)
						{
							e = 0.0;
						}
						else if(e > 1.0)
						{
							e = 1.0;
						}
						unsigned char ec = (unsigned char)(255.0*e);
						if(depthSuperImage[y*superWidth+x] < ec)
						{
							depthSuperImage[y*superWidth+x] = ec;
						}
					}
				}

				// space evenly along curve
				double d = bp->GetDerivt(j, t);
				t+=2.0/(widthFactor+heightFactor)/(0.01+d);
			} while (t<1.000001);
		}
	}

	// overlay paper structure
	bool *superImage = BNEW(bool, superWidth*superHeight);
	for(int i=0; i<superWidth*superHeight; ++i)
	{
		superImage[i] = false;
	}

	for(int x=0; x<superWidth; ++x)
	{
		for(int y=0; y<superHeight; ++y)
		{
			int c = depthSuperImage[y*superWidth+x];
			if(c == 0) // pixel not set
			{
				continue;
			}

			// linear noise
			double s = 0;
			const int numOct = 2;
			{
				double size = 1.0;
				for(int octave = 0; octave < numOct; ++octave, size *= 2.0)
				{
					double tx = x/structureSizeSuper*size;
					double ty = y/structureSizeSuper*size;

					double c00 = 7656.567*sin(5343.937+13.523*sqrt(floor(tx)  )+floor(ty)  );
					c00 -= floor(c00);
					double c01 = 7656.567*sin(5343.937+13.523*sqrt(floor(tx)  )+floor(ty)+1);
					c01 -= floor(c01);
					double c10 = 7656.567*sin(5343.937+13.523*sqrt(floor(tx)+1)+floor(ty)  );
					c10 -= floor(c10);
					double c11 = 7656.567*sin(5343.937+13.523*sqrt(floor(tx)+1)+floor(ty)+1);
					c11 -= floor(c11);

					double fx = tx - floor(tx);
					double fy = ty - floor(ty);
					double gx = 1.0 - fx;
					double gy = 1.0 - fy;

					s += (c00*gx*gy + c01*gx*fy + c10*fx*gy + c11*fx*fy)/size;
				}
			}
			s /= 2.0 - pow(0.5, numOct-1); // bring to range 0...1

			if(c >= 255.0*s)
			{
				superImage[y*superWidth+x] = true;
			}
		}
	}

	delete depthSuperImage;

	// supersampling
	for(int x=0; x<bm.GetWidth(); ++x)
	{
		for(int y=0; y<bm.GetHeight(); ++y)
		{
			int n = 0; // becomes sum of gray values
			for(int xd=0; xd<superSamp; ++xd)
			{
				for(int yd=0; yd<superSamp; ++yd)
				{
					if(superImage[(y*superSamp+yd)*superWidth + x*superSamp+xd])
					{
						++n;
					}
				}
			}
			// 1.0 - ...: paint black, not white
			double val = pow(1.0-n/(double)(superSamp*superSamp), 1.0/gamma);
			// no gamma needed for blending with former value because we are multiplying
			int intVal = (int)(0.5+255.0*val);
			for(int c=0; c<3; ++c)
			{
				const int a = 255;
				const int b = a/2;
				bm.SetPixel(x, y, c, (unsigned char)((intVal*bm.GetPixel(x, y, c) + b)/a));
			}
		}
	}

	BDELETE(superImage);
	BDELETE(clmpParams);
}